import sys
from array import array
from contextlib import contextmanager
from types import SimpleNamespace
from typing import Iterator
from unittest.mock import patch

from pytest import raises

from textual_image._terminal import TerminalError, capture_terminal_response, get_cell_size

if sys.version_info >= (3, 12):
    IntArray = array[int]
else:
    IntArray = array


def test_get_cell_size_stdout_closed() -> None:
    with patch("sys.__stdout__", None):
        with raises(TerminalError):
            get_cell_size()


def test_get_cell_size_on_tty_ioctl_success() -> None:
    if hasattr(get_cell_size, "_result"):
        delattr(get_cell_size, "_result")

    with patch("sys.__stdout__.isatty", return_value=True):
        with patch("textual_image._terminal.get_tiocgwinsz", return_value=(58, 120, 960, 928)):
            assert get_cell_size() == (8, 16)


def test_get_cell_size_on_tty_escape_sequence() -> None:
    if hasattr(get_cell_size, "_result"):
        delattr(get_cell_size, "_result")

    @contextmanager
    def capture_terminal_response(
        start_marker: str, end_marker: str, timeout: float | None = None
    ) -> Iterator[SimpleNamespace]:
        yield SimpleNamespace(sequence="\x1b[6;8;16t")

    with patch("sys.__stdout__"):
        with patch("textual_image._terminal.get_tiocgwinsz", side_effect=OSError()):
            with patch("textual_image._terminal.capture_terminal_response", capture_terminal_response):
                get_cell_size()


def test_get_cell_size_on_tty_failure() -> None:
    if hasattr(get_cell_size, "_result"):
        delattr(get_cell_size, "_result")

    @contextmanager
    def capture_terminal_response(
        start_marker: str, end_marker: str, timeout: float | None = None
    ) -> Iterator[SimpleNamespace]:
        raise TimeoutError()

    with patch("sys.__stdout__"):
        with patch("textual_image._terminal.get_tiocgwinsz", side_effect=OSError()):
            with patch("textual_image._terminal.capture_terminal_response", capture_terminal_response):
                with raises(TerminalError):
                    get_cell_size()


def test_get_cell_size_stdout_not_a_tty() -> None:
    if hasattr(get_cell_size, "_result"):
        delattr(get_cell_size, "_result")

    with patch("sys.__stdout__.isatty", return_value=False):
        term_size = get_cell_size()

    assert term_size.width == 10
    assert term_size.height == 20


def test_capture_terminal_response_stdin_closed() -> None:
    with patch("sys.__stdin__", None):
        with raises(TerminalError):
            with capture_terminal_response("[S]", "[E]"):
                pass


def test_capture_terminal_response_success() -> None:
    with patch("sys.__stdin__"):
        with patch("textual_image._terminal.capture_mode"):
            with patch("textual_image._terminal.read", side_effect="[S]message[E]"):
                with capture_terminal_response("[S]", "[E]") as response:
                    pass

            assert response.sequence == "[S]message[E]"


def test_capture_terminal_response_timeout() -> None:
    with patch("sys.__stdin__"):
        with patch("textual_image._terminal.capture_mode"):
            with patch("textual_image._terminal.read", side_effect=TimeoutError()):
                with raises(TimeoutError):
                    with capture_terminal_response("[S]", "[E]"):
                        pass


def test_capture_terminal_response_unexpected_response() -> None:
    with patch("sys.__stdin__"):
        with patch("textual_image._terminal.capture_mode"):
            with patch("textual_image._terminal.read", side_effect="Something unexpected"):
                with raises(TerminalError):
                    with capture_terminal_response("[S]", "[E]"):
                        pass



from itertools import groupby

from PIL import Image as PILImage

from textual_image._utils import grouped

DCS = "\x1bP"
ST = "\x1b\\"
COLORS = 256


def image_to_sixels(image: PILImage.Image) -> str:
    """Convert an image to Sixels.

    This function convert a `PIL.Image.Image` into the Sixel format.

    Args:
        image: Image to convert to Sixels

    Returns:
        The `image` as Sixel data.
    """
    image = image.convert("RGB").convert("P", palette=PILImage.Palette.ADAPTIVE, colors=COLORS)

    return "".join(
        [
            _get_header(image),
            _get_body(image),
            _get_terminator(),
        ]
    )


def _get_header(image: PILImage.Image) -> str:
    # See <https://vt100.net/docs/vt3xx-gp/chapter14.html#S14.2.1> for details about the parameters.
    sixel_mode = f"{DCS}0;0;0"
    raster_attributes = f'q"1;1;{image.width};{image.height}'

    color_registers = []
    for i, color in enumerate(grouped(image.getpalette() or [], 3)):
        color_str = ";".join(str(int(channel / 256 * 100)) for channel in color)
        color_registers.append(f"#{i};2;{color_str}")

    return f'{sixel_mode}{raster_attributes}{"".join(color_registers)}'


def _get_body(image: PILImage.Image) -> str:
    tokens = []

    for y, row in enumerate(grouped(image.getdata(), image.width)):
        n = 1 << y % 6
        for color, group in groupby(row):
            tokens.append(f"#{color}")

            count = len(list(group))
            if count < 3:
                tokens.append(chr(0x3F + n) * count)
            else:
                tokens.append(f"!{count}{chr(0x3F + n)}")

        tokens.append("-" if n == 32 else "$")

    return "".join(tokens)


def _get_terminator() -> str:
    return ST



"""Functionality to interact with the terminal."""

import logging
import sys
from contextlib import contextmanager
from types import SimpleNamespace
from typing import Iterator, NamedTuple, cast

# pragma: no cover: start -- platform specific, we always will only execute one branch
if sys.platform == "win32":
    from textual_image._win32 import capture_mode, get_tiocgwinsz, read
else:
    from textual_image._posix import capture_mode, get_tiocgwinsz, read
# pragma: no cover: stop

logger = logging.getLogger(__name__)


class TerminalError(Exception):
    """Error thrown on failing terminal operations."""

    pass


class CellSize(NamedTuple):
    """Size of terminal cells."""

    width: int
    """Width of a terminal cell in pixels."""
    height: int
    """Height of a terminal cell in pixels."""


def get_cell_size() -> CellSize:
    """Get size information from the terminal.

    This function is querying the terminal only once. For any call after the first, a cached result is returned.

    Returns:
        The size information

    """
    if not sys.__stdout__:
        raise TerminalError("stdout is closed")

    if hasattr(get_cell_size, "_result"):
        return cast(CellSize, get_cell_size._result)

    width = 0
    height = 0

    if sys.__stdout__.isatty():
        # Try to get the cell size via ioctl
        try:
            rows, columns, screen_width, screen_height = get_tiocgwinsz()
            width = int(screen_width / columns)
            height = int(screen_height / rows)
        except OSError:
            logger.debug("Failed to get cell size via ioctl, falling back to escape sequence")

    if sys.__stdout__.isatty() and (height == 0 or width == 0):
        # Didn't work, let's try to do it via escape sequence
        try:
            with capture_terminal_response("\x1b[", "t", 0.1) as response:
                sys.__stdout__.write("\x1b[16t")
                sys.__stdout__.flush()

            sequence = response.sequence[len("\x1b[") : -len("t")]
            _, height, width = [int(v) for v in sequence.split(";")]
        except (TerminalError, TimeoutError) as e:
            raise TerminalError("Failed to get cell size") from e

    if not sys.__stdout__.isatty():
        logger.debug("Not connected to a terminal, assuming VT340 sizes")
        width = 10
        height = 20

    cell_size = CellSize(width, height)
    setattr(get_cell_size, "_result", cell_size)
    return cell_size


@contextmanager
def capture_terminal_response(
    start_marker: str, end_marker: str, timeout: float | None = None
) -> Iterator[SimpleNamespace]:
    """Captures a terminal response.

    Captures the terminal's response to an escape sequence.
    This is a bit flaky -- keystrokes during reading the response can lead to false answers.
    Additionally, when the terminal does *not* doesn't send an answer, the first character
    of stdin may get lost as this function reads it to determine if it is the response.
    Anyway, as this is improbable to happen, it should be fine.

    Please not this function will not work anymore once Textual is started. Textual runs a threads to read stdin
    and will grab the response.

    Args:
        start_marker: The start sequence of the expected response
        end_marker: The end sequence of the expected response
        timeout: The number of seconds to wait for the response. None to disable timeout.

    Returns:
        The terminal's response
    """
    if not sys.__stdin__:
        raise TerminalError("stdin is closed")

    response = SimpleNamespace(sequence="")

    stdin = sys.__stdin__.buffer.fileno()

    with capture_mode():
        yield response

        while not response.sequence.endswith(end_marker):
            response.sequence += read(stdin, 1, timeout)

            if not response.sequence.startswith(start_marker[: len(response.sequence)]):
                raise TerminalError("Unexpected response from terminal")



from PIL import Image as PILImage
from syrupy.assertion import SnapshotAssertion

from tests.data import TEST_IMAGE
from textual_image._sixel import image_to_sixels


def test_image_to_sixels(snapshot: SnapshotAssertion) -> None:
    with PILImage.open(TEST_IMAGE) as image:
        image = image.resize((16, 16))

    assert image_to_sixels(image) == snapshot



# pragma: exclude file -- this is platform specific code, either this or _posix.py won't get tested,
# depending on the platform the tests are running

import ctypes
import msvcrt
import os
import sys
from contextlib import contextmanager
from ctypes.wintypes import DWORD
from typing import Iterator, Tuple

kernel32 = ctypes.WinDLL("kernel32")  # type: ignore


class Win32TerminalError(Exception):
    """Error thrown on failing terminal operations."""

    pass


def get_tiocgwinsz() -> Tuple[int, int, int, int]:
    raise OSError("TIOCGWINSZ not supported on Windows")


def read(fd: int, length: int, timeout: float | None = None) -> str:
    if timeout is not None:
        wait_for_object(fd, timeout)
    return os.read(fd, length).decode()


@contextmanager
def capture_mode() -> Iterator[None]:
    if not sys.__stdin__:
        raise Win32TerminalError("stdout is closed")

    stdin = sys.stdin.buffer.fileno()

    original_mode = get_console_mode(stdin)
    set_console_mode(stdin, 0x0200)  # ENABLE_VIRTUAL_TERMINAL_INPUT

    try:
        flush(stdin)
        yield
    finally:
        set_console_mode(stdin, original_mode)


def wait_for_object(fd: int, timeout: float) -> None:
    handle = msvcrt.get_osfhandle(fd)  # type: ignore
    ret: int = kernel32.WaitForSingleObject(handle, DWORD(int(timeout * 1000)))
    if ret == 0x00000102:  # WAIT_TIMEOUT
        raise TimeoutError("Timeout waiting for data")


def get_console_mode(fd: int) -> int:
    handle = msvcrt.get_osfhandle(fd)  # type: ignore
    mode = DWORD()
    kernel32.GetConsoleMode(handle, ctypes.byref(mode))
    return mode.value


def set_console_mode(fd: int, mode: int) -> None:
    handle = msvcrt.get_osfhandle(fd)  # type: ignore
    kernel32.SetConsoleMode(handle, mode)


def flush(fd: int) -> None:
    handle = msvcrt.get_osfhandle(fd)  # type: ignore
    kernel32.FlushConsoleInputBuffer(handle)



from importlib import import_module, reload
from unittest import skipIf, skipUnless
from unittest.mock import patch

from tests.data import TEXTUAL_ENABLED


@skipUnless(TEXTUAL_ENABLED, "Textual support disabled")
def test_main() -> None:
    with patch("sys.argv", ["unittest", "rich"]):
        with patch("textual_image.demo.renderable.run") as run_rich:
            with patch("textual_image.demo.widget.run") as run_textual:
                main = import_module("textual_image.__main__")
    assert run_rich.called
    assert not run_textual.called

    with patch("sys.argv", ["unittest", "textual"]):
        with patch("textual_image.demo.renderable.run") as run_rich:
            with patch("textual_image.demo.widget.run") as run_textual:
                reload(main)
    assert not run_rich.called
    assert run_textual.called

    with patch("sys.argv", ["unittest", "textual"]):
        with patch("textual_image.demo.renderable.run") as run_rich:
            with patch("textual_image.demo.widget.run") as run_textual:
                with patch("importlib.util.find_spec", return_value=None):
                    reload(main)
    assert not run_rich.called
    assert not run_textual.called


@skipIf(TEXTUAL_ENABLED, "Textual support enabled")
def test_main_rich_only() -> None:
    with patch("sys.argv", ["unittest", "rich"]):
        with patch("textual_image.demo.renderable.run") as run_rich:
            main = import_module("textual_image.__main__")
    assert run_rich.called

    with patch("sys.argv", ["unittest", "textual"]):
        with patch("textual_image.demo.renderable.run") as run_rich:
            with patch("sys.stderr") as stderr:
                reload(main)
    assert stderr.write.called
    assert not run_rich.called



# pragma: exclude file -- this is platform specific code, either this or _win32.py won't get tested,
# depending on the platform the tests are running

import os
import sys
import termios
import tty
from array import array
from contextlib import contextmanager
from fcntl import ioctl  # type: ignore
from select import select
from typing import Iterator, Tuple


class PosixTerminalError(Exception):
    """Error thrown on failing terminal operations."""

    pass


def get_tiocgwinsz() -> Tuple[int, int, int, int]:
    if not sys.__stdout__:
        raise PosixTerminalError("stdout is closed")

    buf = array("H", [0, 0, 0, 0])
    ioctl(sys.__stdout__, termios.TIOCGWINSZ, buf)  # type: ignore
    rows, columns, screen_width, screen_height = buf
    return rows, columns, screen_width, screen_height


def read(fd: int, length: int, timeout: float | None = None) -> str:
    readable, _, _ = select([fd], [], [], timeout)
    if not readable:
        raise TimeoutError("Timeout waiting for data")
    return os.read(fd, length).decode()


@contextmanager
def capture_mode() -> Iterator[None]:
    if not sys.__stdin__:
        raise PosixTerminalError("stdout is closed")

    stdin = sys.__stdin__.buffer.fileno()
    original_mode = termios.tcgetattr(stdin)  # type: ignore
    tty.setcbreak(stdin, termios.TCSANOW)  # type: ignore

    try:
        yield
    finally:
        termios.tcsetattr(stdin, termios.TCSANOW, original_mode)  # type: ignore



#!/usr/bin/env python

"""Run the textual_image demo."""

import sys
from argparse import ArgumentParser
from importlib.util import find_spec

from textual_image.demo.renderable import RENDERING_METHODS

textual_available = bool(find_spec("textual"))
default_mode = "rich" if find_spec("textual") is None else "textual"

parser = ArgumentParser(description="Demo the capabilities of textual-image")
parser.add_argument("mode", choices=["rich", "textual"], nargs="?", default=default_mode)
parser.add_argument("-m", "--method", choices=RENDERING_METHODS.keys(), default="auto")
arguments = parser.parse_args()

if arguments.mode == "rich":
    from textual_image.demo.renderable import run as run_rich_demo

    run_rich_demo(arguments.method)
elif not textual_available:
    sys.stderr.write(
        "Optional Textual dependency not available. Install this package as `textual-image[textual]` for Textual support."
    )
else:
    from textual_image.demo.widget import run as run_textual_demo

    run_textual_demo(arguments.method)
